@@ -442,11 +442,10 @@ ipset_parse_proto(struct ipset_session *session,
 	protoent = getprotobyname(strcasecmp(str, "icmpv6") == 0
 				  ? "ipv6-icmp" : str);
 	if (protoent == NULL) {
-		uint8_t protonum;
-		int err;
+		uint8_t protonum = 0;
 
-		if (!((err = string_to_u8(session, str, &protonum) == 0) &&
-		      (protoent = getprotobynumber(protonum)) != NULL))
+		if (string_to_u8(session, str, &protonum) ||
+		    (protoent = getprotobynumber(protonum)) == NULL)
 			return syntax_err("cannot parse '%s' "
 					  "as a protocol", str);
 	}
@@ -615,7 +614,9 @@ ipset_parse_proto_port(struct ipset_session *session,
 		default:
 			if (!STREQ(a, "0")) {
 				syntax_err("Protocol %s can be used "
-					   "with pseudo port value 0 only.");
+					   "with pseudo port value 0 only.",
+					   tmp);
+				err = -1;
 				goto error;
 			}
 			ipset_data_flags_set(data, IPSET_FLAG(opt));
@@ -1477,6 +1478,34 @@ ipset_parse_after(struct ipset_session *session,
 	return ipset_data_set(data, opt, str);
 }
 
+/**
+ * ipset_parse_uint64 - parse string as an unsigned long integer
+ * @session: session structure
+ * @opt: option kind of the data
+ * @str: string to parse
+ *
+ * Parse string as an unsigned long integer number.
+ * The value is stored in the data blob of the session.
+ *
+ * Returns 0 on success or a negative error code.
+ */
+int
+ipset_parse_uint64(struct ipset_session *session,
+		   enum ipset_opt opt, const char *str)
+{
+	unsigned long long value = 0;
+	int err;
+
+	assert(session);
+	assert(str);
+
+	err = string_to_number_ll(session, str, 0, ULLONG_MAX - 1, &value);
+	if (err)
+		return err;
+
+	return ipset_session_data_set(session, opt, &value);
+}
+
 /**
  * ipset_parse_uint32 - parse string as an unsigned integer
  * @session: session structure
